"""Types for the pyvconf package."""

# from configparser import SectionProxy

ItemName = str
ItemValue = object
# If we define Items in types.py as SectionProxy, then
# a dict(ConfigParser().items()) will match a Sections type. Once the
# .mypy_cache is updated, we can revert Items to a dict, and the aforementioned
# type check will still pass. Or, we could just ignore the error for now.
# Items = SectionProxy
Items = dict[ItemName, ItemValue]
SectionName = str
Sections = dict[SectionName, Items]




"""An abstract config parser."""

# pylint: disable=too-few-public-methods

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

Config = TypeVar("Config")


class Parser(ABC, Generic[Config]):
    """Interface to parse a config source."""

    @abstractmethod
    def parse(self, src: Config) -> None:
        """Parse a config source."""




"""An abstract parser for configs with sections."""

from dataclasses import dataclass, field
from typing import TypeVar

from _pyvconf.parser import Parser
from _pyvconf.types import Sections

Config = TypeVar("Config")


# mypy gives a false positive for dataclasses with abstract methods.
# As a workaround, we could split the dataclass and abstract portions into
# their own classes, then compose the Parser from these two classes. See
# https://github.com/python/mypy/issues/5374#issuecomment-568335302. But,
# we will instead just suppress the error.
@dataclass  # type: ignore
class SectionParser(Parser[Config]):
    """Interface to parse a config source with sections."""
    sections: Sections = field(default_factory=dict)




"""A concrete INI file parser."""

# pylint: disable=too-few-public-methods

import configparser

from _pyvconf.sectionparser import SectionParser


class IniParser(SectionParser[str]):
    """Parses an INI file."""

    def parse(self, src: str) -> None:
        """Parse an INI file."""
        config = configparser.ConfigParser()
        config.read(src)
        # If we define Items in types.py as SectionProxy, then the next
        # line passes type checks after .mypy_cache is updated. Then we can
        # revert Items to being a dict, and the next line will continue to
        # pass type checks. However, we will just ignore the error for now.
        self.sections = dict(config.items())  # type: ignore




"""Unit tests for the Parser abstract class."""

# pylint: disable=abstract-class-instantiated
# pylint: disable=too-few-public-methods

import pytest

from pyvconf import Parser


class TestParser:
    """Encapsulates the unit tests for the Parser class."""

    @staticmethod
    def test_instantiation_fails() -> None:
        """Test can't instantiate abstract base class."""
        with pytest.raises(TypeError) as excinfo:
            Parser()  # type: ignore
        assert "Can't instantiate abstract class" in str(excinfo.value)




"""Unit tests for the IniParser class."""

import configparser
import logging
import os
import tempfile
from typing import Generator, Tuple

from _pytest.fixtures import SubRequest
import pytest

from pyvconf import IniParser

IniPath = str
SectionName = str
ItemName = str
ItemValue = str
Items = dict[ItemName, ItemValue]
IniDict = dict[SectionName, Items]
IniPathDictPair = Tuple[IniPath, IniDict]


class TestIniParser:
    """Encapsulates the unit tests for the IniParser class."""

    config_default = {
        "str_val": "1",
        "int_val": "1",
        "float_val": "1",
        "bool_val": "1",
    }
    config_section1 = {
        "str_val": "2",
        "float_val": "1.5",
    }

    @staticmethod
    @pytest.fixture(scope="class", autouse=True)
    def logger() -> Generator[logging.Logger, None, None]:
        """A debugging logger."""
        res = logging.getLogger(__name__)
        res.setLevel(logging.DEBUG)
        file_handler = logging.FileHandler(filename="pytest.log",
                                           encoding="utf-8")
        file_handler.setFormatter(logging.Formatter(
            "%(asctime)s %(levelname)s %(message)s"))
        res.addHandler(file_handler)
        res.debug("test started")
        yield res
        res.debug("test finished\n**********")

    @staticmethod
    @pytest.fixture(scope="class", params=[
        {
            "DEFAULT": {
                "str_val": "1",
                "int_val": "1",
                "float_val": "1",
                "bool_val": "1",
            },
            "section1": {
                "str_val": "2",
                "float_val": "1.5",
            },
        },
        {
            "DEFAULT": {
                "str_val": "2",
                "float_val": "1.5",
            },
            "section1": {
                "str_val": "1",
                "int_val": "1",
                "float_val": "1",
                "bool_val": "1",
            },
        },
        {
            "DEFAULT": {
                "str_val": "2",
                "float_val": "1.5",
            },
            "section1": {
                "str_val": "1",
                "int_val": "1",
                "float_val": "1",
                "bool_val": "1",
            },
            "section2": {
                "int_val": "1",
                "float_val": "1",
                "bool_val": "1",
            },
        },
    ])
    def ini_file(request: SubRequest,
                 logger: logging.Logger) -> Generator[IniPathDictPair,
                                                      None, None]:
        """Create an ini file and return its path."""
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as ini_fp:
            config = configparser.ConfigParser()
            for section, items in request.param.items():
                config[section] = items
            config.write(ini_fp)
        logger.debug(ini_fp.name)
        yield (ini_fp.name, request.param)
        os.remove(ini_fp.name)

    @staticmethod
    def test_parse(ini_file: IniPathDictPair, logger: logging.Logger) -> None:
        """Test parsing an INI file."""
        ini_path, ini_dict = ini_file
        parser = IniParser()
        parser.parse(ini_path)

        assert parser.sections.keys() == ini_dict.keys()

        default = ini_dict["DEFAULT"]
        for ini_section, ini_items in ini_dict.items():
            full_ini_items = {**default, **ini_items}
            parser_items = parser.sections[ini_section]
            logger.debug(ini_section)
            logger.debug("    ini_items     : %s", ini_items)
            logger.debug("    full_ini_items: %s", full_ini_items)
            logger.debug("    parser_items  : %s", dict(parser_items))
            assert parser_items == full_ini_items




"""pyvconf interface module."""

from _pyvconf.iniparser import IniParser
from _pyvconf.parser import Parser
from _pyvconf.sectionparser import SectionParser
from _pyvconf.types import Items, Sections

__all__ = [
    "IniParser",
    "Items",
    "Parser",
    "SectionParser",
    "Sections",
]

